/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>
#include "vector.h"
#include "vty.h"
#include "command.h"
#include "getopt.h"
#include "thread.h"
#include "stream.h"
#include "network.h"
#include "prefix.h"
#include "version.h"
#include "memory.h"
#include "if.h"
#include "sockunion.h"
#include "log.h"
#include "pim.h"
#include "pimd.h"
#include "pim_fsm.h"
#include <string.h>
#include <netinet/igmp.h>
#include <linux/mroute.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "assert.h"
#include "tables.h"
#include "pim_packet.h"
#include "igmp_def.h"
#include "zbug.h"
#include "pim_route.h"
#include "if_idx_nexthop.h"
#include "pim_assert.h"

ZBUG_EXTERN

char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */
extern int pim_sock, igmp_sock;
extern struct best_route best_route;



list mg_list;
list mrt_hash[MRT_HASH_SIZE];
list mgrp_hash[MGRP_HASH_SIZE];


void tables_init( void )
{
  mg_list = list_init();
  memset(mrt_hash, 0, MRT_HASH_SIZE * sizeof(  list ));
  memset(mgrp_hash, 0, MGRP_HASH_SIZE * sizeof( list ));
}


/** Functions for manipulation of the mrt hash. **/


struct mrt *mr_lookup(u_int32_t src, u_int32_t group )
{
  struct mrt *mr;
  listnode node;
  list l;
  u_int32_t idx = (src^group) % MRT_HASH_SIZE;

  l = mrt_hash[idx];
  if(!l)
    return NULL;

  for (node=listhead (l); node; nextnode (node)) {  
    mr = getdata(node);
    if (mr->ms->src == src && mr->mg->group == group)
      return mr;
  }
  return NULL;
}

int
entry_timeout_thread(struct thread *t_entry_timeout)
{
  struct mrt *mr = THREAD_ARG (t_entry_timeout);
  time_t now;
  
  zassert(mr, 0);
  if( mr == NULL) return 0;
  mr->t_entry_timeout = NULL;
  
  printf("Removing (%s, %s) \n", inet_2a( mr->ms->src, b1), 
	 inet_2a( mr->mg->group, b2));
  ZBUG_G(mr->mg->group)
    {
      time(&now);
      fprintf(ZBUG_LOG, "entry data timeout (%s,%s) %s",
	      inet_2a(mr->ms->src, b1),
	      inet_2a(mr->mg->group, b2),
	      ctime(&now));
      pim_flags_f(ZBUG_LOG, mr->ms->src, mr->mg->group, mr->flags);
      fflush(ZBUG_LOG);
    }
  mr_remove(mr);
  return 1;
}

int
data_timer_thread(struct thread *t_entry_timeout)
{
  u_int32_t timer;
  struct mgrp *mg;
  struct mrt *mr;
  listnode node1, node2;

  timer = DATA_TIMER_INTERVAL;
  t_pim_data =  thread_add_timer(master, data_timer_thread, NULL, timer);
  
  for(node1=listhead (mg_list); node1; nextnode (node1))
    {
      mg=getdata(node1);

      for(node2=listhead (mg->mrt_list); node2; nextnode (node2))
	{
	  mr=getdata(node2);
	  get_sg_counts(mr);
	  
	  if(mr->stats.packets != mr->stats.old_packets && 
	     !(mr->flags & MRT_RPT))
#define DATA_TIMEOUT 210 /* See pp. 39 */
	    mr_entry_timer_update(mr, DATA_TIMEOUT ); /* Data timeout */
	  
	}
    }
  return 0;
}

struct mrt *mr_add(struct msrc *ms, struct mgrp *mg )
{
  u_int32_t idx = (ms->src^mg->group) % MRT_HASH_SIZE;
  list l;
  time_t now;
  struct mrt *mr;

  mr = XMALLOC(MTYPE_PIM_SG, sizeof(struct mrt));
  bzero (mr, sizeof (struct mrt));

  ZBUG_G(mg->group)
    {
      time(&now);
      fprintf(ZBUG_LOG, "Creating %s,%s (%u) %s",
	      inet_2a(ms->src, b1),
	      inet_2a(mg->group, b2),
	      mg->mrt_list->count,
	      ctime(&now));
      fflush(ZBUG_LOG);
    }
  l = mrt_hash[idx];
  if(!l) {
    l = list_init();
    mrt_hash[idx] = l;
  }
  ms->use_count++;
  mr->oif_list = list_init();
  mr->ms = ms;
  mr->mg = mg;
  mr->uptime = time(NULL);
  mr->exptime = time(NULL);
  mr->t_entry_timeout = NULL;
  mr->flags = 0;
  list_add_node(l, mr);

#if 1
  {
    /* Add sorted into list */
    listnode node, prev_node = 0;
    for(node=listhead(mg->mrt_list); node; nextnode(node))
      {
	if( ntohl(((struct mrt *)getdata(node))->ms->src) > ntohl(mr->ms->src))
	  break;
	prev_node = node;
      }
    
    if(node)
      list_add_node_prev(mg->mrt_list, node, mr);
    else
      {
	if(prev_node)
	  list_add_node_next(mg->mrt_list, prev_node, mr);
	else
	  list_add_node(mg->mrt_list, mr);
      }
  }  
#else
  list_add_node(mg->mrt_list, mr);
#endif
  return mr;
}

int mr_remove(struct mrt *mr)
{
  struct mrt *p;
  struct mgrp *mg;
  struct oif *mi;
  listnode node;
  time_t now;
  u_int32_t found;
  u_int32_t idx = (mr->ms->src^mr->mg->group) % MRT_HASH_SIZE;
  vifbitmap_t oifs;

  if( mr == NULL) return 0;
  
  mg = mr->mg;

  ZBUG_G(mg->group)
    {
      time(&now);
      fprintf(ZBUG_LOG, "Removing %s,%s %s",
	      inet_2a(mr->ms->src, b1),
	      inet_2a(mg->group, b2),
	      ctime(&now));
      fflush(ZBUG_LOG);  
    }
  
  oifs = mr_oifs(mr);
  mfc_delete(igmp_sock, mr->ms->src, mr->mg->group, mr_iif(mr), oifs);
  
  for (node=listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node); 
    if(mi->igmp.t_igmp_timeout)  thread_cancel(mi->igmp.t_igmp_timeout);
    if(mi->pim.t_oif_timeout)  thread_cancel(mi->pim.t_oif_timeout);
    if(mi->igmp.t_send_group_specific_query)  
      thread_cancel(mi->igmp.t_send_group_specific_query);
    XFREE(MTYPE_PIM_OIF, mi);
  }
  list_delete_all(mr->oif_list);
  
  /* Remove us from grp list and stop pending timers */

  if(mr->t_entry_timeout) {
    thread_cancel(mr->t_entry_timeout);
    mr->t_entry_timeout = NULL;
  }

  if(mr->t_register_stop) {
    thread_cancel(mr->t_register_stop);
    mr->t_register_stop = NULL;
  }

  /* Find check and unlink entry */
  
  for (found =0, node=listhead (mg->mrt_list); node; nextnode (node)) {
    p = getdata (node);
    if( mr == p) {
      found = 1;
      break;
    }
  }
  if(mrt_hash[idx])
    {
      list_delete_by_val(mrt_hash[idx], mr);
      list_free_ifempty( mrt_hash[idx] );
    }
  
  mr->ms->use_count--;  
  XFREE(MTYPE_PIM_SG, mr);
  
  if(! found) return -1;

  list_delete_node(mg->mrt_list, node);
  return 1;
}


/** Functions for manipulation of the grp hash. **/

struct mgrp *mg_lookup(u_int32_t G )
{
  struct mgrp *mg;
  u_int32_t idx = G % MGRP_HASH_SIZE;
  listnode node;
  list l;
  
  l = mgrp_hash[idx];
  
  if(!l)
    return NULL;
  
  for (node=listhead (l); node; nextnode (node)) {  
    mg = getdata(node);
    if (mg->group == G) {
      return mg;
    }
  }
  return NULL;
}

struct mgrp *mg_add(u_int32_t G )
{
  u_int32_t idx = G % MGRP_HASH_SIZE;
  list l;
  struct mgrp *mg = XMALLOC(MTYPE_PIM_G, sizeof(struct mgrp));
  bzero (mg, sizeof (struct mgrp));

  l = mgrp_hash[idx];
  if(!l) {
    l = list_init();
    mgrp_hash[idx] = l;
  }
  
  mg->group = G;
  mg->oif_list = list_init ();
  mg->mrt_list = list_init ();

  list_add_node(l, mg);

  {
    /* Add group sorted into list */
    listnode node, prev_node = 0;
    for(node=listhead(mg_list); node; nextnode(node))
      {
	if( ntohl(((struct mgrp *)getdata(node))->group) > ntohl(mg->group))
	  break;
	prev_node = node;
      }
    
    if(node)
      list_add_node_prev(mg_list, node, mg);
    else
      {
	if(prev_node)
	  list_add_node_next(mg_list, prev_node, mg);
	else
	  list_add_node(mg_list, mg);
      }
  }
  
  ZBUG_G(mg->group)
    {
      fprintf(ZBUG_LOG, "Creating *,%s\n",
	      inet_2a(mg->group, b2));
      fflush(ZBUG_LOG);
    }
  return mg;
}

int mg_remove( struct mgrp *mg )
{

  struct mrt *mr;
  struct oif *mi;
  listnode node;
  u_int32_t idx = mg->group % MGRP_HASH_SIZE;

  printf("Removing group %s\n", inet_2a(mg->group, b1));
  ZBUG_G(mg->group)
    {
      fprintf(ZBUG_LOG, "Removing *,%s (%u) ",
	      inet_2a(mg->group, b2), mg->mrt_list->count);
      pim_flags_f(ZBUG_LOG, 0, mg->group, mg->flags);
      fflush(ZBUG_LOG);
    }
  
  //mg->pim.rp->use_count--;

  /* Remove all cache entries too before removing the list */
  
  if(mg->mrt_list != NULL) {
    while(!list_isempty(mg->mrt_list))
      {
	mr = getdata( listhead( mg->mrt_list));
	mr_remove( mr );
      }
    list_delete_all(mg->mrt_list);
  }
  
  if(mg->oif_list != NULL) {
    for (node=listhead (mg->oif_list); node; nextnode (node)) {
      mi = getdata (node); 
      if(mi->igmp.t_igmp_timeout)  thread_cancel(mi->igmp.t_igmp_timeout);
      if(mi->pim.t_oif_timeout)  thread_cancel(mi->pim.t_oif_timeout);
      if(mi->igmp.t_send_group_specific_query)  
	thread_cancel(mi->igmp.t_send_group_specific_query);
      XFREE(MTYPE_PIM_OIF, mi);
    }
    list_delete_all(mg->oif_list);
  }

  if(mgrp_hash[idx])
    {
      list_delete_by_val(mgrp_hash[idx], mg);
      list_free_ifempty( mgrp_hash[idx] );
    }

  list_delete_by_val(mg_list, mg);
  
  XFREE(MTYPE_PIM_G, mg);
  return 1;
}

/** Functions for manipulation of the src hash. **/

list msrc_hash[MSRC_HASH_SIZE];

struct msrc *ms_lookup(u_int32_t S )
{
  struct msrc *s;
  list l;
  listnode node;
  u_int32_t idx = S % MSRC_HASH_SIZE;
  
  l = msrc_hash[idx];
  if(!l)
    return NULL;
  
  for(node = listhead(l);node;nextnode(node))
    {
      s = getdata(node);
      zassert(s, NULL);
      if(s->src == S)
	return s;
    }
  return NULL;
}

struct msrc *ms_add(u_int32_t S)
{
  u_int32_t idx = S % MSRC_HASH_SIZE;
  list l;
  struct msrc *s;
  u_int32_t iif, rpf_neighbor;
  struct pim_interface *pi;
  struct interface *ifp;

  s = ms_lookup( S );
  if(s)
    return s;
  
  iif = if_idx_nexthop((struct in_addr *) &S);
  if(iif == -1)
    return 0;
  
  rpf_neighbor =  best_route.gw;
  ifp = get_iflist_ifp(iif);
  if(ifp == 0)
    return 0;

  s = XMALLOC(MTYPE_PIM_SRC, sizeof(struct msrc));
  bzero (s, sizeof (struct msrc));
  s->src = S;

  s->rpf_neighbor =  rpf_neighbor;
  s->ifp = ifp;

  pi = s->ifp->info;

  /* For now we can only detect connected sources. Cisco and gated 
     uses 0 as preference and so do we...
     */

  if( s->rpf_neighbor == 0) {
    s->preference = 0;
    s->metric = 0;
  }
  else {
    s->preference = ASSERT_NOTFOUND_PREFERENCE;
    s->metric = ASSERT_NOTFOUND_METRIC;
  }

  s->uptime = time(NULL);
  s->use_count = 0;

  if(!(l = msrc_hash[idx]))
    l = msrc_hash[idx] = list_init();

  list_add_node(l, s);
  return s;
}

int ms_remove(u_int32_t S )
{
  struct msrc *s;
  u_int32_t idx = S % MSRC_HASH_SIZE;
  
  s =  (struct msrc *) ms_lookup(S);
  if( s == NULL)
    return 0;
  list_delete_by_val( msrc_hash[idx], s);
  
  list_free_ifempty( msrc_hash[idx] );

  XFREE(MTYPE_PIM_SRC, s);
  return 1;
}

struct oif *
oif_alloc (int memtype)
{
  struct oif *mi;

  mi = XMALLOC (memtype, sizeof (struct oif)); 
  bzero (mi, sizeof (struct oif));

return mi;
}

/* Remove an interface from the mgrp-list AND If this was the last member 
   we remove the group entry as well */

int 
mg_oif_list_delete (struct oif *mi)
{
  struct mgrp *mg;

  zassert(mi, -1);
  mg = mi->mg;
  zassert(mg, -1);
  
  ZBUG_G(mg->group)
    {
      fprintf(ZBUG_LOG, "mg_oif_delete\nGroup stat: ");
      pim_flags_f(ZBUG_LOG, 0, mg->group, mg->flags);
      fprintf(ZBUG_LOG, "Oif stat: ");
      pim_flags_f(ZBUG_LOG, 0, mg->group, mi->flags);
      fflush(ZBUG_LOG);
    }
  
  list_delete_by_val(mg->oif_list, mi);
  
  if( mi->pim.t_oif_timeout) 
    thread_cancel(mi->pim.t_oif_timeout);
  if( mi->igmp.t_igmp_timeout) 
    thread_cancel(mi->igmp.t_igmp_timeout);
  if( mi->igmp.t_send_group_specific_query) 
    thread_cancel(mi->igmp.t_send_group_specific_query);
  XFREE(MTYPE_PIM_OIF, mi);
  return 1;
}

int 
mr_oif_list_delete (struct oif *mi)
{
  struct mrt *mr;

  zassert(mi, -1);
  mr = mi->mr;
  
  ZBUG_G(mr->mg->group)
    {
      fprintf(ZBUG_LOG, "mr_oif_delete\nEntry stat: ");
      pim_flags_f(ZBUG_LOG, mr->ms->src, mr->mg->group, mr->flags);
      fprintf(ZBUG_LOG, "Oif stat: ");
      pim_flags_f(ZBUG_LOG, mr->ms->src, mr->mg->group, mi->flags);
      fflush(ZBUG_LOG);
    }

  if( mi->pim.t_oif_timeout) 
    thread_cancel(mi->pim.t_oif_timeout);
  if( mi->igmp.t_igmp_timeout) 
    thread_cancel(mi->igmp.t_igmp_timeout);
  if( mi->igmp.t_send_group_specific_query) 
    thread_cancel(mi->igmp.t_send_group_specific_query);

  list_delete_by_val(mr->oif_list, mi);  

  XFREE(MTYPE_PIM_OIF, mi);
  return 1;
}

struct oif *
mr_oif_lookup(struct mrt *mr, struct interface *ifp)
{

 struct oif *mi;
  listnode node;

  if( !mr->oif_list || list_isempty(mr->oif_list)) return 0;

  for (node = listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) return mi;
  }
  return NULL;
}

struct oif *
mg_oif_lookup(struct mgrp *mg, struct interface *ifp)
{

  struct oif *mi;
  listnode node;

  if( !mg->oif_list || list_isempty(mg->oif_list)) return 0;

  for (node = listhead (mg->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) return mi;
  }
  return NULL;
}

int
mr_oif_delete(struct mrt *mr, struct interface *ifp)
{

 struct oif *mi;
  listnode node;

  if( !mr->oif_list || list_isempty(mr->oif_list)) return 0;

  for (node = listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) {
      mr_oif_list_delete(mi);
      break;
    }
  }
  return 1;
}

void 
mg_pim_timer_update(struct mgrp *mg, time_t newtime)
{
  if(mg->pim.exptime < time(NULL) + newtime) {
    if( mg->pim.t_entry_timeout) thread_cancel(mg->pim.t_entry_timeout);
    mg->pim.exptime = time(NULL)+newtime;
    mg->pim.t_entry_timeout = thread_add_timer(master, rt_1_wc_1_timeout, mg, 
					   newtime);
  }
  else {
    printf("YYYY  skipping mgrp pim timer update for %s %lu/%lu YYYYY\n"
    ,inet_2a(mg->group, b1), mg->pim.exptime, newtime);
  }
}


void 
mr_entry_timer_update(struct mrt *mr, time_t newtime)
{
  if(mr->exptime < time(NULL) + newtime) {
    if( mr->t_entry_timeout) thread_cancel(mr->t_entry_timeout);
    mr->exptime = time(NULL)+newtime;
    mr->t_entry_timeout = thread_add_timer(master, entry_timeout_thread, mr, 
					   newtime);
  }
}

u_int32_t mg_iif( struct mgrp *mg)
{
  /* Asserts does not matter here, ifp is always the same. */  
  
  return mg->pim.rp->ifp->ifindex;
}

struct interface *mr_iif_ifp( struct mrt *mr)
{
  if(mr->flags & (MRT_SPT | MRT_REGISTER))
    return mr->ms->ifp;
  
  /* Asserts does not matter here, ifp is always the same. */
  
  return mr->mg->pim.rp->ifp;
}

u_int32_t mr_iif( struct mrt *mr)
{
  return mr_iif_ifp(mr)->ifindex;
}

vifbitmap_t
mgrp_oifs(struct mgrp *mg)
{
  struct oif *mi;
  vifbitmap_t oifs = 0;
  listnode node;

  for (node = listhead (mg->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    if(!(mi->flags & MRT_ASSERTED))
      VIFM_SET(mi->ifp->ifindex, oifs);
  }

  if((mg->flags & MRT_REGISTER) && !(mg->flags & MRT_REGISTER_STOP ))
      VIFM_SET(PIMREG_IDX, oifs);

  return oifs;
}

vifbitmap_t mr_oifs(struct mrt *mr)
{
  struct oif *mi;
  vifbitmap_t oifs = 0;
  listnode node;

  for (node = listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    if(!(mi->flags & MRT_ASSERTED))
      VIFM_SET(mi->ifp->ifindex, oifs);
  }

  if((mr->flags & MRT_REGISTER) &&
     !(mr->flags & MRT_REGISTER_STOP ))
    {
      mi = mg_oif_lookup(mr->mg, mr_iif_ifp(mr));
      if(!(mi && mi->flags & MRT_ASSERTED))
	VIFM_SET(PIMREG_IDX, oifs);
    }
  
  /*     VIFM_ISSET(mr_iif(mr), mgrp_oifs(mr->mg)))
	 fel !(mr->mg->flags & MRT_ASSERTED))
	 VIFM_SET(PIMREG_IDX, oifs); */
  
  return oifs;
}

void
copy_mg_oifs(struct mrt *mr, struct mgrp *mg)
{
  listnode node;
  struct oif *mi, *mi_tmp;
  
  if(!list_isempty(mr->oif_list))
    printf("list not empty\n");
  
  if(!mr->oif_list) 
    mr->oif_list = list_init();
  
  for (node = listhead (mg->oif_list); node; nextnode (node)) {
    mi_tmp  = getdata(node);

    if((mr->flags & MRT_SPT) && mr->ms->ifp == mi_tmp->ifp) continue;
    else if(mg->pim.rp->ifp == mi_tmp->ifp) continue;
    
    mi = XMALLOC(MTYPE_PIM_OIF, sizeof( struct oif));
    memcpy(mi, getdata (node), sizeof( struct oif));
    mi->igmp.t_igmp_timeout = NULL;
    mi->pim.t_oif_timeout = NULL;
    mi->igmp.t_send_group_specific_query = NULL;
    mi->pim.exptime = 0;
    mi->uptime = time(NULL);
    mi->flags &= ~MRT_IGMP_IN_USE;
    mi->mr = mr;

    list_add_node(mr->oif_list, mi);
  }
}

void
add_oif_to_mr(struct mgrp *mg, struct interface *ifp)
{
  listnode node, node2;
  struct oif *mi, *mi_tmp;
  struct mrt *mr;
  vifbitmap_t oifs = 0;
  u_int32_t idx;

  for (node = listhead (mg->mrt_list); node; nextnode (node)) {
    mr = getdata(node);
    
    /* SPT-bit handling. Asserts? */


    /* (c) ii */
    if(mr->flags & MRT_SPT)
      if( ifp == mr->ms->ifp && same_subnet(mr->ms->src)) 
	continue;

    /* (c) iii */
    if(!(mr->flags & MRT_SPT))
      {
	/* When RPT iif is towards the RP */
	if( ifp == mr->mg->pim.rp->ifp )
	  continue;
      }
    else
      /* When SPT iif is towards source */
      if( ifp == mr->ms->ifp ) 
	continue;

    for (node2 = listhead (mr->oif_list); node2; nextnode (node2)) {
      mi_tmp =getdata(node2);
      if( ifp == mi_tmp->ifp ) return; /* We got this alraedy */
    }

    mi = oif_alloc(MTYPE_PIM_OIF);
    mi->ifp = ifp;
    mi->igmp.t_igmp_timeout = NULL;
    mi->pim.t_oif_timeout = NULL;
    mi->igmp.t_send_group_specific_query = NULL;
    mi->pim.exptime = 0;
    mi->uptime = time(NULL);
    mi->flags &= ~MRT_IGMP_IN_USE;
    mi->mr = mr;

    list_add_node(mr->oif_list, mi);

    oifs = mr_oifs(mr);

    if(mr->flags & MRT_SPT) idx = mr->ms->ifp->ifindex;
    else idx =  mg->pim.rp->ifp->ifindex;

      mfc_add(igmp_sock,
	      mr->ms->src,
	      mr->mg->group,
	      idx,
	      oifs);

  }
}

void sync_mg_family(struct mgrp *mg)
{
  listnode node, oif_node;
  struct oif *mi;
  struct mrt *mr;

//  printf("XXXXX--- SYNC ----XXXXX\n");
  
  /* For every oif on this (*,G) */
  for (node = listhead (mg->oif_list); node; nextnode (node))
    {
      mi = getdata(node);
      if((mg->flags & MRT_IGMP_IN_USE) && (mg->flags & MRT_PIM_IN_USE))
	{
	  mi->flags &= ~MRT_BLACKHOLE;
	  continue;
	}
      
      if(mg->flags & MRT_IGMP_IN_USE)
	{
	  mi->flags &= ~MRT_BLACKHOLE;
	  continue;
	} 

      if(mg->flags & MRT_PIM_IN_USE)
	{
	  mi->flags &= ~MRT_BLACKHOLE;
	  continue;
	}
    }

  /* For every (S,G) on this (*,G) */
  for (node = listhead (mg->mrt_list); node; nextnode (node))
    {
      mr = getdata(node);
      if((mg->flags & MRT_IGMP_IN_USE) && (mg->flags & MRT_PIM_IN_USE))
	{
	  mr->flags &= ~MRT_BLACKHOLE;
	  mr->flags |= MRT_PIM_IN_USE;
	  goto oif_loop;
	}
      
      if(mg->flags & MRT_IGMP_IN_USE)
	{
	  mr->flags &= ~MRT_BLACKHOLE;
	  mr->flags &= ~MRT_PIM_IN_USE;
	  goto oif_loop;
	} 
      
      if(mg->flags & MRT_PIM_IN_USE)
	{
	  mr->flags &= ~MRT_BLACKHOLE;
	  mr->flags |= MRT_PIM_IN_USE;
	  mr->flags &= ~MRT_IGMP_IN_USE;
	  goto oif_loop;
	}
      
oif_loop:      
      /* For every oif on this (S,G) */
      for (oif_node = listhead (mr->oif_list); oif_node; nextnode (oif_node))
	{
	  mi = getdata(oif_node);
	  if((mg->flags & MRT_IGMP_IN_USE) && (mg->flags & MRT_PIM_IN_USE))
	    {
	      mi->flags |= MRT_PIM_IN_USE;
	      continue;
	    }
	  
	  if(mg->flags & MRT_IGMP_IN_USE)
	    {
	      mi->flags &= ~MRT_PIM_IN_USE;
	      continue;
	    } 
	  
	  if(mg->flags & MRT_PIM_IN_USE)
	    {
	      mi->flags |= MRT_PIM_IN_USE;
	      mi->flags &= ~MRT_IGMP_IN_USE;
	      continue;
	    }
	}
    }
}

void mr_add_oif(struct mrt *mr, struct interface *ifp)
{
  struct oif *mi;

    mi = oif_alloc(MTYPE_PIM_OIF);
    if(!ifp)
      {
	*(int *)0 = 1;
	exit(1);
      }
    mi->ifp = ifp;
    mi->igmp.t_igmp_timeout = NULL;
    mi->pim.t_oif_timeout = NULL;
    mi->igmp.t_send_group_specific_query = NULL;
    mi->pim.exptime = 0;
    mi->uptime = time(NULL);
    mi->flags &= ~MRT_IGMP_IN_USE;
    mi->mr = mr;
    list_add_node(mr->oif_list, mi);
}

void ms_timeout(void)
{
  struct msrc *s;
  listnode node;
  int i;
  
  for(i=0;i <  MSRC_HASH_SIZE; i++) {
    if( msrc_hash[i] == 0)  continue;
    for (node = listhead(msrc_hash[i]); node; nextnode(node)) {
      s = getdata( node );

      /* We timeout only one source per hash  each time */

      if(s && s->use_count == 0) {
	/*	printf("Source %s is timed out\n", inet_2a(s->src, b1)); */
	ms_remove(s->src);
	break;
      }
    }
  }
}

int ms_timeout_thread(struct thread *t_ms_timeout)
{
  ms_timeout();
   t_ms_timeout = 
     thread_add_timer(master, ms_timeout_thread, NULL, MS_PEROIDIC_TIMER ); 
  return 1;
}


int
mr_flag_oif(struct mrt *mr, struct interface *ifp, u_int32_t flag)
{
  listnode node;
  struct oif *mi;

  for (node = listhead (mr->oif_list); node; nextnode (node))
    {
      mi = getdata(node);
      if(mi->ifp == ifp)
	{
	  mi->flags |= flag;
	  return 1;
	}
    }
  return 0;
}

int
mg_flag_oif(struct mgrp *mg, struct interface *ifp, u_int32_t flag)
{
  listnode node;
  struct mrt *mr;
  struct oif *mi;
  
  for (node = listhead (mg->oif_list); node; nextnode (node))
    {
      mi = getdata(node);
      if(mi->ifp == ifp)
	{
	  mi->flags |= flag;
	  break;
	}
    }
  
  for (node = listhead (mg->mrt_list); node; nextnode (node))
    {
      mr = getdata(node);
      mr_flag_oif(mr, ifp, flag);
    }
  
  return 0;
}

int mr_mfc_add( struct mrt *mr )
{
  vifbitmap_t oifs = mr_oifs(mr);

  return mfc_add(igmp_sock,
		 (u_int32_t) mr->ms->src,
		 (u_int32_t) mr->mg->group,
		 mr_iif(mr),
		 oifs);
}

int mg_mfc_add( struct mgrp *mg )
{
  listnode node;
  struct mrt *mr;
  
  for (node = listhead (mg->mrt_list); node; nextnode (node))
    {
      mr = getdata(node);
      mr_mfc_add(mr);
    }
  return 1;
}


u_int32_t mg_rpf_neighbor( struct mgrp *mg )
{
  if(mg->flags & MRT_ASSERTED)
    return mg->pim.asserted.rpf_neighbor;
  
  return mg->pim.rp->rpf_neighbor;
}

u_int32_t mr_rpf_neighbor( struct mrt *mr )
{
  if(mr->flags & MRT_ASSERTED)
    return mr->asserted.rpf_neighbor;
  
  return mr->ms->rpf_neighbor;
}
